@@ -1,4 +1,5 @@
 import json
+import time
 from collections import namedtuple
 from jsonpath_ng.ext import parse
 from ixnetwork_restpy import SessionAssistant
@@ -46,11 +47,8 @@ class IxNetworkApi(Api):
     def config(self):
         return self._config
 
-    @property
-    def config_objects(self):
-        """A dict of all unique names to abstract config objects
-        """
-        return self._config_objects
+    def get_config_object(self, name):
+        return self._config_objects[name]
 
     @property
     def ixn_objects(self):
@@ -68,21 +66,34 @@ class IxNetworkApi(Api):
     def assistant(self):
         return self._assistant
 
+    def _dict_to_obj(self, source):
+        """Returns an object given a dict
+        """
+        if isinstance(source, list):
+            source = [self._dict_to_obj(x) for x in source]
+        if not isinstance(source, dict):
+            return source
+        o = lambda : None
+        for k, v in source.items():
+            o.__dict__[k] = self._dict_to_obj(v)
+        return o
+
     def set_config(self, config):
         """Abstract API implementation
         """
         if isinstance(config, (Config, str, dict, type(None))) is False:
             raise TypeError('The content must be of type (Config, str, dict, type(None))' % Config.__class__)
         if isinstance(config, str) is True:
-            config = json.loads(config, object_hook = lambda otg : namedtuple('otg', otg.keys()) (*otg.values())) 
+            config = self._dict_to_obj(json.loads(config)) 
         elif isinstance(config, dict) is True:
-            config = namedtuple('otg', config.keys())(*config.values())
+            config = self._dict_to_obj(config)
         self._config = config
         self._config_objects = {}
         self._ixn_objects = {}
+        self._capture_request = None
         self._errors = []
         self.validation.validate_config()        
-        self.__connect()
+        self._connect()
         if self._config is None:
             self._ixnetwork.NewConfig()
         else:
@@ -96,6 +107,42 @@ class IxNetworkApi(Api):
         """
         return self.traffic_item.transmit(request)
 
+    def set_port_capture(self, request):
+        """Starts capture on all ports that have capture enabled.
+        """
+        self._capture_request = request
+
+    def _start_capture(self):    
+        if self._capture_request is not None:   
+            payload = { 'arg1': [] }
+            for name in self._capture_request.port_names:
+                payload['arg1'].append(self._vport.find(Name=name).href)
+            url = '%s/vport/operations/clearCaptureInfos' % self._ixnetwork.href
+            response = self._request('POST', url, payload)
+            self._ixnetwork.StartCapture()
+
+    def get_capture_results(self, request):
+        """Gets capture file and returns it as a byte stream
+        """
+        capture = self._vport.find(Name=request.port_name).Capture
+        capture.Stop('allTraffic')
+        time.sleep(3)
+
+        payload = { 'arg1': [self._vport.href] }
+        url = '%s/vport/operations/getCaptureInfos' % self._ixnetwork.href
+        response = self._request('POST', url, payload)
+        file_name = response['result'][0]['arg6']
+        file_id = response['result'][0]['arg1']
+
+        url = '%s/vport/operations/saveCaptureInfo' % self._ixnetwork.href
+        payload = { 'arg1': self._vport.href, 'arg2': file_id }
+        response = self._request('POST', url, payload)
+
+        path = '%s/capture' % self._ixnetwork.Globals.PersistencePath
+        url = '%s/files?absolute=%s&filename=%s.cap' % (self._ixnetwork.href, path, file_name)
+        pcap_file_bytes = self._request('GET', url)
+        return pcap_file_bytes
+
     def get_port_results(self, request):
         """Abstract API implementation
         """
@@ -103,8 +150,25 @@ class IxNetworkApi(Api):
 
     def get_flow_results(self, request):
         """Abstract API implementation
+
+        Args
+        ----
+        - request (Union[FlowRequest, str, dict]): A request for flow results.
+            The request content MUST be based on the OpenAPI #/components/schemas/Result.FlowRequest model.
+            See the docs/openapi.yaml document for all model details.
         """
-        return self.traffic_item.results(request)
+        from abstract_open_traffic_generator.result import FlowRequest
+        self._errors = []
+        if isinstance(request, (FlowRequest, str, dict)) is False:
+            raise TypeError('The content must be of type Union[FlowRequest, str, dict]')
+        if isinstance(request, str) is True:
+            request = json.loads(request, object_hook = lambda otg : namedtuple('otg', otg.keys()) (*otg.values())) 
+        elif isinstance(request, dict) is True:
+            request = namedtuple('otg', request.keys())(*request.values())
+        response = self.traffic_item.results(request)
+        if len(self._errors) > 0:
+            raise Exception('\n'.join(self._errors))
+        return response
 
     def add_error(self, error):
         """Add an error to the global errors
@@ -114,7 +178,7 @@ class IxNetworkApi(Api):
         else:
             self._errors.append(error)
 
-    def __connect(self):
+    def _connect(self):
         """Connect to an IxNetwork API Server.
         """
         if self._assistant is None:
@@ -129,6 +193,26 @@ class IxNetworkApi(Api):
             self._traffic = self._ixnetwork.Traffic
             self._traffic_item = self._ixnetwork.Traffic.TrafficItem
 
+    def _request(self, method, url, payload=None):
+        connection, url = self._assistant.Session._connection._normalize_url(url)
+        headers = {}
+        if payload is not None:
+            payload = json.dumps(payload)
+            headers['Content-Type'] = 'application/json'
+        response = self._assistant.Session._connection._session.request(method, url, headers=headers, data=payload, verify=False)
+        response.raise_for_status()
+        if response.status_code == 202:
+            content = response.json()
+            while content['state'] == 'IN_PROGRESS':
+                time.sleep(1)
+                response = self._request('GET', content['url'])
+        if response.headers.get('Content-Type'):
+            if response.headers['Content-Type'] == 'application/json':
+                return response.json()
+            elif response.headers['Content-Type'] == 'application/octet-stream':
+                return response.content
+        return response
+
     def _remove(self, ixn_obj, items):
         """Remove any ixnetwork items that are not found in the configuration list.
         If the list does not exist remove everything.
@@ -141,6 +225,9 @@ class IxNetworkApi(Api):
         else:
             ixn_obj.find().remove()
 
+    def _get_topology_name(self, port_name):
+        return 'Topology %s' % port_name
+        
     def select_vports(self):
         """Select all vports.
         Return them in a dict keyed by vport name.
@@ -169,7 +256,7 @@ class IxNetworkApi(Api):
                 vports[vport['name']] = vport
         return vports
 
-    def select_traffic_items(self, filters=[]):
+    def select_traffic_items(self, traffic_item_filters=[]):
         """Select all traffic items.
         Return them in a dict keyed by traffic item name.
 
@@ -187,12 +274,12 @@ class IxNetworkApi(Api):
                         {
                             'child': 'trafficItem',
                             'properties': ['name', 'state', 'enabled'],
-                            'filters': filters
+                            'filters': traffic_item_filters
                         },
                         {
                             'child': 'highLevelStream',
                             'properties': ['txPortName', 'rxPortNames'],
-                            'filters': filters
+                            'filters': []
                         }
                     ],
                     'inlines': []
