@@ -1,5 +1,7 @@
 import json
 from ixnetwork_open_traffic_generator.customfield import CustomField
+from ixnetwork_restpy import StatViewAssistant
+
 
 class TrafficItem(CustomField):
     """TrafficItem configuration
@@ -32,6 +34,8 @@ class TrafficItem(CustomField):
         'pfcPause': 'pfc_pause',
         'vlan': 'vlan',
         'ipv4': 'ipv4',
+        "tcp" : "tcp",
+        "udp" : "udp",
         'custom': 'custom'
     }
 
@@ -40,6 +44,8 @@ class TrafficItem(CustomField):
         'pfcpause': 'pfcPause',
         'vlan': 'vlan',
         'ipv4': 'ipv4',
+        "tcp" : "tcp",
+        "udp" : "udp",
         'custom': 'custom'
     }
 
@@ -86,6 +92,38 @@ class TrafficItem(CustomField):
         'priority' : '_ipv4_priority',
     }
 
+    _TOS = {
+        "precedence": "ipv4.header.priority.tos.precedence",
+        "delay": "ipv4.header.priority.tos.delay",
+        "throughput": "ipv4.header.priority.tos.throughput",
+        "reliability": "ipv4.header.priority.tos.reliability",
+        "monetary": "ipv4.header.priority.tos.monetary",
+        "unused": "ipv4.header.priority.tos.unused"
+    }
+    
+    _TCP = {
+        "src_port" : "tcp.header.srcPort",
+        "dst_port" : "tcp.header.dstPort",
+        "ecn_ns" : "tcp.header.ecn.nsBit",
+        "ecn_cwr" : "tcp.header.ecn.cwrBit",
+        "ecn_echo" : "tcp.header.ecn.ecnEchoBit",
+        "ctl_urg" : "tcp.header.controlBits.urgBit",
+        "ctl_ack" : "tcp.header.controlBits.ackBit",
+        "ctl_psh" : "tcp.header.controlBits.pshBit",
+        "ctl_rst" : "tcp.header.controlBits.rstBit",
+        "ctl_syn" : "tcp.header.controlBits.synBit",
+        "ctl_fin" : "tcp.header.controlBits.finBit",
+    }
+    
+    _UDP = {
+        "src_port" : "udp.header.srcPort",
+        "dst_port" : "udp.header.dstPort",
+        "length" : "udp.header.length",
+        "checksum" : "udp.header.checksum",
+    }
+    
+    _CUSTOM = '_custom_headers'
+    
     def __init__(self, ixnetworkapi):
         self._api = ixnetworkapi
         
@@ -104,34 +142,42 @@ class TrafficItem(CustomField):
             for flow in self._api.config.flows:
                 args = {
                     'Name': flow.name,
-                    'TrafficItemType': 'l2L3'
+                    'TrafficItemType': 'l2L3',
+                    'TrafficType': self._get_traffic_type(flow)
                 }
-                if flow.endpoint is None:
-                    raise ValueError('%s endpoint cannot be None' % flow.name)
-                elif flow.endpoint.choice == 'port':
-                    args['TrafficType'] = 'raw'
-                elif flow.endpoint.device.packet_encap in ['ethernet', 'vlan']:
-                    args['TrafficType'] = 'ethernetVlan'
-                else:
-                    args['TrafficType'] = flow.endpoint.device.packet_encap
                 ixn_traffic_item.find(Name=flow.name)
                 if len(ixn_traffic_item) == 0:
                     ixn_traffic_item.add(**args)
                 else:
                     ixn_traffic_item.update(**args)
-                self._configure_endpoint(ixn_traffic_item.EndpointSet, flow.endpoint)
-                if flow.enable is not None:
-                    ixn_traffic_item.Enabled = flow.enable
+                self._configure_endpoint(ixn_traffic_item.EndpointSet, flow.tx_rx)
                 ixn_traffic_item.Tracking.find().TrackBy = ['trackingenabled0']
                 ixn_stream = ixn_traffic_item.ConfigElement.find()
                 self._configure_stack(ixn_stream, flow.packet)
                 self._configure_size(ixn_stream, flow.size)
                 self._configure_rate(ixn_stream, flow.rate)
                 self._configure_tx_control(ixn_stream, flow.duration)
+    
+    def _get_traffic_type(self, flow):
+        if flow.tx_rx is None:
+            raise ValueError('%s Flow.tx_rx property cannot be None' % flow.name)
+        elif flow.tx_rx.choice == 'port':
+            encap = 'raw'
+        else:
+            encap = None
+            for name in flow.tx_rx.device.tx_device_names:
+                device = self._api.get_config_object(name)
+                if device.choice == 'ethernet':
+                    encap = 'ethernetVlan'
+                elif device.choice == 'bgpv4':
+                    encap = 'ipv4'
+                else:
+                    encap = device.choice
+        return encap
 
     def _configure_endpoint(self, ixn_endpoint_set, endpoint):
-        """Transform flow.endpoint to /trafficItem/endpointSet
-        The model allows for only one endpoint per traffic item
+        """Transform flow.tx_rx to /trafficItem/endpointSet
+        The model allows for only one endpointSet per traffic item
         """
         args = {
             'Sources': [],
@@ -156,13 +202,14 @@ class TrafficItem(CustomField):
             ixn_endpoint_set.update(**args)
 
     def _configure_stack(self, ixn_stream, headers):
-        """Transform flow.packets[0..n] to /traffic/trafficItem/configElement/stack 
+        """Transform flow.packets[0..n] to /traffic/trafficItem/configElement/stack
         The len of the headers list is the definitive list which means add/remove
         any stack items so that the stack list matches the headers list.
         If the headers list is empty then use the traffic generator default stack.
         """
         stacks_to_remove = []
         ixn_stack = ixn_stream.Stack.find()
+        headers = self.adjust_header(headers)
         for i in range(0, len(headers)):
             header = headers[i]
             if len(ixn_stack) <= i:
@@ -180,6 +227,7 @@ class TrafficItem(CustomField):
                 else:
                     stack = ixn_stack[i]
             self._configure_field(stack.Field, header)
+        self._api._traffic_item.Generate()
         for stack in stacks_to_remove:
             stack.Remove()
     
@@ -189,18 +237,23 @@ class TrafficItem(CustomField):
         stack_href = ixn_stack.AppendProtocol(template)
         return ixn_stream.Stack.read(stack_href)
 
-    def _configure_field(self, ixn_field, header):
+    def _configure_field(self, ixn_field, header, field_choice=False):
         """Transform flow.packets[0..n].header.choice to /traffic/trafficItem/configElement/stack/field
         """
         field_map = getattr(self, '_%s' % header.choice.upper())
         packet = getattr(header, header.choice)
+        if isinstance(field_map, dict) is False:
+            method = getattr(self, field_map)
+            method(ixn_field, packet)
+            return
+        
         for packet_field_name in dir(packet):
             if packet_field_name in field_map:
                 pattern = getattr(packet, packet_field_name)
                 field_type_id = field_map[packet_field_name]
-                self._configure_pattern(ixn_field, field_type_id, pattern)
+                self._configure_pattern(ixn_field, field_type_id, pattern, field_choice)
 
-    def _configure_pattern(self, ixn_field, field_type_id, pattern):
+    def _configure_pattern(self, ixn_field, field_type_id, pattern, field_choice=False):
         if pattern == None:
             return
         
@@ -211,19 +264,27 @@ class TrafficItem(CustomField):
 
         ixn_field = ixn_field.find(FieldTypeId=field_type_id)
         if pattern.choice == 'fixed':
-            ixn_field.update(Auto=False, ValueType='singleValue', SingleValue=pattern.fixed)
+            ixn_field.update(Auto=False,
+                ActiveFieldChoice=field_choice,
+                ValueType='singleValue',
+                SingleValue=pattern.fixed)
         elif pattern.choice == 'list':
-            ixn_field.update(Auto=False, ValueType='valueList', ValueList=pattern.list)
+            ixn_field.update(Auto=False,
+                ActiveFieldChoice=field_choice,
+                ValueType='valueList',
+                ValueList=pattern.list)
         elif pattern.choice == 'counter':
             value_type = 'increment' if pattern.counter.up is True else 'decrement'
-            ixn_field.update(Auto=False, 
-                ValueType=value_type, 
+            ixn_field.update(Auto=False,
+                ValueType=value_type,
+                ActiveFieldChoice=field_choice,
                 StartValue=pattern.counter.start,
                 StepValue=pattern.counter.step,
                 CountValue=pattern.counter.count)
         elif pattern.choice == 'random':
-            ixn_field.update(Auto=False, 
-                ValueType='repeatableRandomRange', 
+            ixn_field.update(Auto=False,
+                ActiveFieldChoice=field_choice,
+                ValueType='repeatableRandomRange',
                 MinValue=pattern.random.min,
                 MaxValue=pattern.random.max,
                 StepValue=pattern.random.step,
@@ -288,7 +349,7 @@ class TrafficItem(CustomField):
                     StartDelay=duration.fixed.delay,
                     StartDelayUnits=duration.fixed.delay_unit)
             else:
-                ixn_tx_control.update(Type='fixedFrameCount', 
+                ixn_tx_control.update(Type='fixedFrameCount',
                     FrameCount=duration.fixed.packets,
                     MinGapBytes=duration.fixed.gap,
                     StartDelay=duration.fixed.delay,
@@ -314,13 +375,16 @@ class TrafficItem(CustomField):
         3) Execute requested transmit action (start|stop|pause|resume)
         """
         regex = None
-        if request.names is not None and len(request.names) > 0:
-            regex = '^(%s)$' % '|'.join(request.names)
-        self._api._traffic_item.find(regex)
+        if request.flow_names is not None and len(request.flow_names) > 0:
+            regex = '^(%s)$' % '|'.join(request.flow_names)
         if request.state == 'start':
+            self._api._traffic_item.find(Name=regex, State='unapplied')
             self._api._ixnetwork.StartAllProtocols('sync')
-            self._api._traffic_item.Generate()
-            self._api._traffic.Apply()
+            if len(self._api._traffic_item) > 0:
+                self._api._traffic_item.Generate()
+                self._api._traffic.Apply()
+            self._api._start_capture()
+        self._api._traffic_item.find(Name=regex)
         if request.state == 'start':
             self._api._traffic_item.StartStatelessTrafficBlocking()
         elif request.state == 'stop':
@@ -336,8 +400,14 @@ class TrafficItem(CustomField):
     def results(self, request):
         """Return flow results
         """
+        filter = {
+            'property': 'name',
+            'regex': '.*'
+        }
+        if request is not None and request.flow_names is not None and len(request.flow_names) > 0:
+            filter['regex'] = '^(%s)$' % '|'.join(request.flow_names)
         flow_rows = {}
-        for traffic_item in self._api.select_traffic_items().values():
+        for traffic_item in self._api.select_traffic_items(traffic_item_filters=[filter]).values():
             flow_row = [0 for i in range(len(TrafficItem._RESULT_COLUMNS))]
             self._set_result_value(flow_row, 'name', traffic_item['name'])
             self._set_result_value(flow_row, 'state', traffic_item['state'])
@@ -346,6 +416,7 @@ class TrafficItem(CustomField):
             flow_rows[traffic_item['name']] = flow_row
         try:
             table = self._api.assistant.StatViewAssistant('Traffic Item Statistics')
+            table.AddRowFilter('Traffic Item', StatViewAssistant.REGEX, filter['regex'])
             for row in table.Rows:
                 flow_row = flow_rows[row['Traffic Item']]
                 self._set_result_value(flow_row, 'frames_tx', row['Tx Frames'])
@@ -357,7 +428,7 @@ class TrafficItem(CustomField):
                 self._set_result_value(flow_row, 'loss', row['Loss %'])
         except Exception as e:
             self._api.add_error(e)
-        results = { 
+        results = {
             'columns': TrafficItem._RESULT_COLUMNS,
             'rows': flow_rows.values()
         }
